#
# Interface to "load" and "save" Objects to or from Blender when
# running as a native script.
# 

import Blender
from PCD import Object, Scene
from PCD.Maths.Vector3D import Vec3


def B2GLAxes (v):
    "Converts from Blender axes to OpenGL axes"
    return tuple ( [v[0], v[2], -v[1]] )

def B2GLAxesN (v):
    "Converts from Blender axes to OpenGL axes without inverting an axis"
    return tuple ( [v[0], v[2], v[1]] )

def GL2BAxes (v):
    "Converts from OpenGL axes to Blender axes"
    return tuple ( [v[0], -v[2], v[1]] )

def GL2BAxesN (v):
    "Converts from OpenGL axes to Blender axes without inverting an axis"
    return tuple ( [v[0], v[2], v[1]] )




def _FacesFromBlender (bface, hasuv):
    if len (bface.v) == 4:
        if hasuv:
            faces = [ ((bface.v[0].index, bface.v[1].index, bface.v[2].index),
                       (bface.v[0].uv, bface.v[1].uv, bface.v[2].uv)),
                      ((bface.v[0].index, bface.v[2].index, bface.v[3].index),
                       (bface.v[0].uv, bface.v[2].uv, bface.v[3].uv)) ]
        else:
            faces = [ ((bface.v[0].index, bface.v[1].index, bface.v[2].index),
                       None ),
                      ((bface.v[0].index, bface.v[2].index, bface.v[3].index),
                       None ) ]
    elif len (bface.v) == 3:
        if hasuv:
            faces = [ ((bface.v[0].index, bface.v[1].index, bface.v[2].index),
                       (bface.v[0].uv, bface.v[1].uv, bface.v[2].uv)) ]
        else:
            faces = [ ((bface.v[0].index, bface.v[1].index, bface.v[2].index),
                       None ) ]
    else:
        print >>sys.stderr, 'WARNING: face has',        \
                                    len(face.v), 'vertices. Ignoring.'

    return faces



def _ObjectFromBlender (bobj):
    "Returns an Object containing the same data as the Blender.Object bobj"
    assert bobj.getType() == 'Mesh'

    obj = Object.New (bobj.getName())

    obj.transform.setPosition (B2GLAxes (bobj.getLocation ()))
    obj.transform.setRotation (B2GLAxesN (bobj.getEuler ()))
    obj.transform.setScale (B2GLAxesN (bobj.getSize ()))

    bmesh = bobj.getData ()
    for bvert in bmesh.verts:
        pos = B2GLAxes (bvert.co)
        obj.verts.append (Vec3 (pos[0], pos[1], pos[2]))

    faces = []
    for bface in bmesh.faces:
        faces.extend (_FacesFromBlender (bface, bmesh.hasFaceUV()))
    for face in faces:
        obj.addTriIndices (face[0], face[1])

    obj.applyScale ()

    return obj


def _ObjectToBlender (obj, matnamemap):
    "Returns a Blender.Object for the supplied Object"

    bmesh = Blender.NMesh.New (obj.name)
    bmesh.mode = 0    # make everything single sided

    bfaceimg = None
    if obj.material:
        bmat = matnamemap[obj.material.getSaveName()]
        bmesh.addMaterial (bmat)
        bfaceimg = bmat.getTextures()[0].tex.image

    for vert in obj.verts:
        x,y,z = GL2BAxes (vert)
        bvert = Blender.NMesh.Vert (x,y,z)
        bmesh.verts.append (bvert)

    for tri_index,tri in enumerate (obj.tris):
        bverts = [bmesh.verts[i] for i in tri]
        bface = Blender.NMesh.Face (bverts)
        if len (obj.texfaces) == len (obj.tris):
            for uv in obj.texfaces[tri_index]:
                bface.uv.append (uv)
        if bfaceimg != None:
            bface.image = bfaceimg
        bmesh.faces.append (bface)

    bobj = Blender.Object.New ('Mesh', obj.name)
    bobj.link (bmesh)
    bobj.loc = GL2BAxes (obj.transform.pos)
    bobj.rot = list (GL2BAxesN (obj.transform.rot))
    bobj.setSize (GL2BAxesN (obj.transform.sca))

    return bobj



def _LoadBObjects (bobjs, unlink):
    scene = Scene.New ()
    bscene = Blender.Scene.GetCurrent ()

    for bobj in bobjs:
        scene.addObject (_ObjectFromBlender (bobj))
        if unlink:
            bscene.unlink (bobj)

    return scene



def _SaveTextures (textures):
    """Adds the Material.Textures to the Blender scene and returns a name
    to Blender.Texture mapping dictionary"""
    map = {}

    for texture in textures:
        btex = Blender.Texture.New (texture.name)
        map[texture.name] = btex
        try:
            print 'Loading image', texture.filename
            bimg = Blender.Image.Load (texture.filename)
            btex.type = Blender.Texture.Types['IMAGE']
            btex.setImage (bimg)
	except IOError, s:
	    print 'Failed to load texture', texture.name, ':', s

    return map


def _SaveMaterials (materials, texnamemap):
    """Adds the materials to the Blender scene and returns a name to
    Blender.Material mapping dictionary. texnamemap should be a map as
    returned by _SaveTextures()"""

    map = {}
    for material in materials:
        bmat = Blender.Material.New (material.getSaveName ())
        map[material.getSaveName()] = bmat
	bmat.setSpec (0.1)
        textures = material.getTextures ()
        for i,texture in enumerate (textures):
            if texture != None:
                btex = texnamemap[texture.name]
                bmat.setTexture (i, btex, Blender.Texture.TexCo.UV)

    return map


def _SaveObjects (objs, matnamemap):
    "Adds the Objects to the current Blender Scene"
    bscene = Blender.Scene.GetCurrent ()
    for obj in objs:
        bscene.link (_ObjectToBlender (obj, matnamemap))
        


def LoadSelection (unlink=False):
    "Returns a Scene representing the current Blender selection"
    return _LoadBObjects (Blender.Object.GetSelected (), unlink)

def Load (unlink=False):
    "Returns a Scene representing the current Blender scene"
    return _LoadBObjects (Blender.Object.Get (), unlink)
    

def Save (scene):
    "Saves the contents of the Scene into the current Blender scene"
    texnamemap = _SaveTextures (scene.getTextures())
    matnamemap = _SaveMaterials (scene.getMaterials(), texnamemap)
    _SaveObjects (scene.getObjects(), matnamemap)
    


