#!BPY

"""
Name: 'PIP Export'
Blender: 231
Group: 'Export'
"""

from Blender import NMesh,Window,Object,Material,Texture
import sys


def B2GLAxes (v):
    "Converts from Blender axes to OpenGL axes"
    return tuple ( [v[0], v[2], -v[1]] )

def B2GLAxesN (v):
    "Converts from Blender axes to OpenGL axes without inverting an axis"
    return tuple ( [v[0], v[2], v[1]] )


class PIPFileWriter:

	indentsize = 4	#xml indent size

	def __init__ (self):
		# file object to write to
		self.outfile = sys.stderr
		self.indentlevel = 0

	def indent (self):
		self.indentlevel += PIPFileWriter.indentsize

	def unindent (self):
		self.indentlevel -= PIPFileWriter.indentsize
		if self.indentlevel < 0:
			self.indentlevel = 0

	def write (self, s):
		i = ' ' * self.indentlevel
		self.outfile.write (i + s + '\n')

	def write_i (self, s):
		self.write (s)
		self.indent ()

	def u_write (self, s):
		self.unindent ()
		self.write (s)

	def writeList (self, type, items, fun, attribs=None):
		"writes a general list, calling fun for each item"
		if attribs == None:
			self.write_i ('<%slist>' % type)
		else:
			self.write_i ('<%slist %s>' % (type, attribs))
			
		for i in items: fun (i)
		
		self.u_write ('</%slist>' % type)


	def writeTexture (self, tex):
		if tex.type == Texture.Types['IMAGE']:
			img = tex.getImage ()
			self.write ('<texture name="%s" src="%s" />' % 
							(tex.name,img.getFilename ()))

	def writeMaterial (self, mat):
		self.write_i ('<material name="%s">' % mat.name)
		textures = mat.getTextures ()
		for i in range(len(textures)):
			t = textures[i]
			if t:
				self.write ('<texchannel num="%d" texture="%s" />' %
								(i, t.tex.name) )
		self.u_write ('</material>')

	def writeVertex (self, v):
		self.write ("<vtx>%f %f %f</vtx>" % B2GLAxes (v.co) )

	def writeTri (self, face, i=(0,1,2)):
		v = face.v
		text = '%d %d %d' % (v[i[0]].index, v[i[1]].index, v[i[2]].index)
		if self.writetexfaces:
			uv = face.uv
			text += ' %f,%f %f,%f %f,%f' % (
						uv[i[0]][0],uv[i[0]][1], 
						uv[i[1]][0],uv[i[1]][1], 
						uv[i[2]][0],uv[i[2]][1] )
		self.write ('<tri>%s</tri>' % text)

	def writeFace (self, face):
		if len (face.v) == 4:
			self.writeTri (face, (0,1,2))
			self.writeTri (face, (0,2,3))
		elif len (face.v) == 3:
			self.writeTri (face)
		else:
			print >>sys.stderr, 'WARNING: face has', len(face.v), 'vertices. Ignoring.'
	
	def writeTransform (self, loc, rot, size):
		self.write_i ('<transform>')
		self.write ('<position>%f %f %f</position>' % B2GLAxes(loc))
		self.write ('<eulerrot>%f %f %f</eulerrot>' %  B2GLAxesN(rot))
		self.write ('<scale>%f %f %f</scale>' %  B2GLAxesN(size))
		self.u_write ('</transform>')

	def writeObject (self, object):
		if object.getType() == 'Mesh':
			name = object.getName()
			mesh = NMesh.GetRawFromObject (name)
			self.write_i ('<object name="%s">' % name)
			self.writeTransform (object.getLocation(), 
                                    object.getEuler(), object.getSize())
			self.writeList ('vtx', mesh.verts, self.writeVertex)
			
			if mesh.hasFaceUV(): self.writetexfaces = 1
			else: self.writetexfaces = 0
			self.writeList ('tri', mesh.faces, self.writeFace,
								'texinfo="%d"'%self.writetexfaces)

			# save zero or one material references
			mtls = mesh.materials
			if mtls:
				self.write ('<matref material="%s" />' % mtls[0].name)

			self.u_write ('</object>')

	def writeScene (self, outfilename):
		"Writes the current scene to a .pip format file"

		# open the output file
		self.outfile = file (outfilename, "w")

		# write the xml header
		self.write ('<?xml version="1.0"?>')

		# scene object is top-level object
		self.write_i ('<scene>')

		# write the texture list
		self.writeList ('tex', Texture.Get(), self.writeTexture)

		# write the material list
		self.writeList ('mat', Material.Get(), self.writeMaterial)

		# write the list of objects in the scene
		objects = Object.Get()
		for object in objects:
			self.writeObject (object)

		self.u_write ('</scene>')

		self.outfile.close ()



# This callback saves the file
def filename_cb (filename):
	w = PIPFileWriter ()
	w.writeScene (filename)


# open a file selector to get the filename
Window.FileSelector (filename_cb, 'Export')
