#!BPY

"""
Name: 'PIP Import'
Blender: 228
Group: 'Import'
"""

import Blender
from Blender import NMesh,Window,Object,Material,Texture,Image
import sys, math
import xml.parsers.expat


def GL2BAxes (v):
    "Converts from OpenGL axes to Blender axes"
    return [v[0], -v[2], v[1]]

def GL2BAxesN (v):
    "Converts from OpenGL axes to Blender axes without inverting an axis"
    return [v[0], v[2], v[1]]



class PIPParseException(Exception):
	"Exception raised if there's a problem parsing a PIP file"
	pass


class PIPFileReader:
	"Parses a PIP geometry file and creates a scene"

	def __init__ (self):
		# create the parser
		self.parser = xml.parsers.expat.ParserCreate()
		self.parser.StartElementHandler = self._starttag
		self.parser.EndElementHandler = self._endtag
		self.parser.CharacterDataHandler = self._chardata

		# the current element- this is a stack
		self.currel = []
		# the text data for the current element
		self.currdata = ''

		# blender sometimes needs to rename long material names...
		self.matnames = {}

		# the current NMesh
		self.currobject = None
		# the name of the current mesh
		self.currname = ''
		# the current Material
		self.currmaterial = None
		

	def load (self, filename):
		"Loads a .pip file"
		f = file (filename, 'r')
		self.parser.ParseFile (f)
		f.close ()
 
		Blender.Redraw ()


	# these are parser callbacks
	def _starttag (self, tag, attrs):
		self.currel.append ((tag,attrs))

		if tag == 'object':
			assert self.currobject == None
			self.currobject = NMesh.New ()
			self.currobject.mode = 0    # make everything single sided
			self.currname = attrs['name']
			self.currloc = (0,0,0)
			self.currrot = (0,0,0)
			self.currsca = (1,1,1)
		
		elif tag == 'texture': 
			tex = Texture.New (attrs['name'])
			
			try:
				print 'Loading image:', attrs['src']
				img = Image.Load (attrs['src'])
				print '  img:', img
				tex.type = Texture.Types['IMAGE']
				tex.setImage (img)
			except IOError, s:
				print 'Failed to load texture', attrs['src'], ':', s

		elif tag == 'material':
			assert self.currmaterial == None
			self.currmaterial = Material.New (attrs['name'])
			self.matnames[attrs['name']] = self.currmaterial.name
			# set up default material
			self.currmaterial.setSpec (0.1)

		elif tag == 'texchannel':
			tex = Texture.Get (attrs['texture'])
			channel = int (attrs['num'])
			self.currmaterial.setTexture (channel, tex, Texture.TexCo.UV)

		elif tag == 'matref':
			matname = self.matnames[attrs['material']]
			mat = Material.Get (matname)
			self.currobject.addMaterial (mat)

		elif tag == 'trilist':
			self.readtexfaces = attrs.get ('texinfo', 0)

	
	def _endtag (self, tag):
		top = self.currel.pop()
		if top[0] != tag:
			raise PIPParseError, "Unmatched tag: %s (%s found)" % (top,tag)

		attrs = top[1]

		# create object for tag
		if tag == 'vtx':
			x,y,z = GL2BAxes ([float(s) for s in self.currdata.split()])
			v = NMesh.Vert (x, y, z)
			self.currobject.verts.append (v)
			self.currdata = ''

		elif tag == 'tri':
			face = NMesh.Face ()
			data = self.currdata.split()
			self.currdata = ''
			indices = [int(i) for i in data[0:3]]

			for i in indices:
				vert = self.currobject.verts[i]
				face.v.append (vert)

			if self.readtexfaces:
				uvs = data[3:6]
				for uv in uvs:
					u,v = uv.split (',')
					face.uv.append ( (float(u),float(v)) )

			self.currobject.faces.append (face)

		elif tag == 'position':
			self.currloc = GL2BAxes ([float(s) for s in self.currdata.split()])
			self.currdata = ''

		elif tag == 'eulerrot':
			self.currrot = GL2BAxesN ([float(s) for s in self.currdata.split()])
			self.currdata = ''

		elif tag == 'scale':
			self.currsca = GL2BAxesN ([float(s) for s in self.currdata.split()])
			self.currdata = ''
		
		elif tag == 'object':
			# update the faces' image
			mats = self.currobject.materials
			if mats:
				tex = mats[0].getTextures()[0].tex
				if tex and tex.type == Texture.Types.IMAGE:
					for face in self.currobject.faces:
						face.image = tex.image

			o = NMesh.PutRaw (self.currobject, self.currname, 1)
			o.setName (self.currname)
			o.setLocation (self.currloc[0], self.currloc[1], self.currloc[2])
			o.setEuler (self.currrot)
			o.setSize (self.currsca[0], self.currsca[1], self.currsca[2])
			self.currobject = None
			self.currname = ''

		elif tag == 'material':
			self.currmaterial = None
		
			
	def _chardata (self, data):
		self.currdata += data



def LoadFile (filename):
	w = PIPFileReader ()
	w.load (filename)


def LoadWithSelector ():
	# open a file selector to get the filename
	Window.FileSelector (LoadFile, 'Import')
	

# if this code is being run, load a file into the current scene
if __name__ == '__main__':
	LoadWithSelector ()
