#
# Base class for feature generation functions
#

from PCD import Registry
import PCD.Maths.Vector3D


class FuncError (Exception):
    def __init__ (self, s):
        Exception.__init__ (self)
        self.s = s
    def __str__ (self):
        return repr (self.s)



class Func (object):
    """Meant only as a base class.
    Subclass must define member variables:
        - _regpath:    path to register this to, eg '/features/window'
        - _regname:    name to register with, eg 'gothic'
        _ _regsockets: list of tuples: (name,type)
    """
    def __init__ (self):
        self._regpath = ''
        self._regname = ''
        self._regsockets = []

    def getRegName (self):
        "Returns the path+name that this object should be registered to"
        return Registry.JoinPath (self._regpath, self._regname)


    def getRegMeta (self):
        "Returns the list of metadata this object needs to have"
        meta = []
        for socket in self._regsockets:
            meta.append ( ('sockets/'+socket[0], socket[1]) )
        return meta


    def chooseSocketValue (self, name, registry, prand):
        "Returns a RegObject for the named socket"

        globalsocket = False
        for socket in self._regsockets:
            if socket[0] == name:
                if len (socket) >= 3 and socket[2]:
                    globalsocket = True
                
        sockname = self.getRegName()+':sockets/'+name
        meta = registry.getObject (sockname)

        ## TODO: handle meta.value

        if globalsocket:
            reqname = name
        else:
            reqname = sockname
        obj = registry.chooseObject (meta.type, prand, reqname)

        return obj
        

    def socketGenContent (self, socketname, size, registry, 
                                        prand, level, translate=None):
        """Calls chooseSocketValue() to obtain a relevant featgenfunc, then 
        calls that object's generateContent(). The result is then translated
        if _translate_ has been passed, and returned."""
        func = self.chooseSocketValue (socketname, registry, prand)
        obj = func.getObject().generateContent (size, registry, prand, level+1)
        if translate:
            translate = PCD.Maths.Vector3D.MakeVector (translate)
            obj.translate (translate)
        return obj


    def _genQuad (self, obj, v, t=None):
        "Adds a quad to the object"
        if t is None:
            obj.addTri ( (v[0], v[1], v[2]) )
            obj.addTri ( (v[0], v[2], v[3]) )
        else:
            obj.addTri ( (v[0], v[1], v[2]), (t[0], t[1], t[2]) )
            obj.addTri ( (v[0], v[2], v[3]), (t[0], t[2], t[3]) )
            

    def _genQuadStrip (self, obj, v0, v1, t0=None, t1=None):
        """Adds a list of triangles spanning vector lists v0 and v1 to obj.
        (v0[0],v1[0],v1[1]) should be a front-facing triangle."""
        assert len(v0) == len(v1)
        if t0 is not None: 
            assert t1 is not None
            assert len(t0) == len(v0)
            assert len(t1) == len(v1)
            
        for i in range (len (v0)):
            ni = (i+1) % len(v0)
            if t0 is None:
                obj.addTri ( ( v0[i], v1[i], v1[ni] ) )
                obj.addTri ( ( v0[i], v1[ni], v0[ni] ) )
            else:
                obj.addTri ( (v0[i], v1[i], v1[ni]), (t0[i], t1[i], t1[ni]))
                obj.addTri ( (v0[i], v1[ni], v0[ni]), (t0[i], t1[ni], t0[ni]))
            
    def _genQuadStripN (self, obj, v0, v1, t0=None, t1=None):
        """Adds a list of triangles spanning vector lists v0 and v1 to obj.
        (v0[0],v1[0],v1[1]) should be a front-facing triangle.
        Doesn't loop around"""
        assert len(v0) == len (v1)
        if t0 is not None: 
            assert t1 is not None
            assert len(t0) == len(v0)
            assert len(t1) == len(v1)
            
        for i in range (len (v0)-1):
            if t0 is None:
                obj.addTri ( (v0[i], v1[i], v1[i+1]) )
                obj.addTri ( (v0[i], v1[i+1], v0[i+1]) )
            else:
                obj.addTri ( (v0[i], v1[i], v1[i+1]), 
                             (t0[i], t1[i], t1[i+1]))
                obj.addTri ( (v0[i], v1[i+1], v0[i+1]), 
                             (t0[i], t1[i+1], t0[i+1]))
            

    def generateContent (self, size, registry, prand, level):
        """The actual generator function [must be overridden by subclasses].

        This function should generate some geometry and return a [Meta]Object.
        Parameters are:
            -size: a 2-tuple, (width,height). Geometry should be contained
                   within: 0<=x<=width,0<=y<=height, and should fill this
                   volume at z=0
            -registry: the Registry to use
            -prand: The PRand to use
            -level: The hierarchy level of this call. Any subfunctions called
                    should have this argument set to level+1. It should be used
                    when generating PRand request names, so that if the same
                    function gets called at different levels it can behave
                    differently, whilst maintaining uniformity at any given
                    level.
        """
        raise NotImplementedError
    


class SetGenFunc (Func):
    "Base class for feature set generators"

    def generateContent (self, required, registry, prand):
        """Must be implemented by subclass.
        _required_ should be a sequence of tuples: 
                            (featname, sockname, size)
        where featname is the name the feature should have in the resulting
        set, sockname is the named socket of this generator to use to create
        the feature (eg 'door'), and size is the required size.
        """
        raise NotImplementedError


