    
class FeatureSet:
    "Encapsulates a set of named architectural features"

    def __init__ (self):
        self._features = {}
        self._materials = {}
        self.mincellsize = (0,0)
        self.name = ''
        self.rooffunc = None
        self._valid = True
    

    def update (self):
        """Updates internal state
        This should be called after a feature has been added or modified""" 

        # loop through the feature list and update the minimum cell size
        minw = 0
        minh = 0
        for feature in self._features.values():
            o = feature.getObject ()

            w = o.getWidth() + feature.getHPad()
            if w > minw: minw = w
            h = o.getHeight() + feature.getVPad()
            if h > minh: minh = h
        
        self.mincellsize = (minw, minh)

        self._valid = True


    def addFeature (self, name, feature):
        "Adds a feature to the set"
        assert name not in self._features
        self._features[name] = feature
        self._valid = False
        

    def getFeature (self, name):
        "Returns a copy of a named feature"
        try:
            return self._features[name].clone ()
        except KeyError:
            return None

    def setRoofFunc (self, rooffunc):
        self.rooffunc = rooffunc        
        
    def roofFunc (self, face, registry, prand):
        "Returns a roof object"
        assert self.rooffunc != None
        if not self._valid:
            self.update ()
        return self.rooffunc.generateContent (face, registry, prand)

    def getMinCellSize (self):
        "Returns the minimum cell size as tuple: (w,h)"
        if not self._valid:
            self.update ()
        return self.mincellsize

    def getMaterials (self):
        "Returns the texture list"
        return self._materials.values()

    def getMaterial (self, name):
        "Returns a named texture or None"
        try: return self._materials[name]
        except KeyError: return None



        
def New ():
    "Returns a new FeatureSet object"
    return FeatureSet ()

    
