import FeatGen
from PCD import Object, MetaObject, Registry, Maths
from PCD.Maths.Vector3D import Vec3
import math, random

# building face types
FRONTFACE, BACKFACE, MISCFACE = range(3)


class Style (FeatGen.Func):
    "Encapsulates an architectural style"

    def __init__ (self):
        self._regpath = '/styles'
        self._regsockets = [ ('MAT_WALLFACE', '/materials/wall', True) ]
    

    def _getFeatureList (self):
        """Subclasses must override this to inform the base class about what
        kinds of features are needed. It should return a sequence of tuples:
            (featurename, class, size)
        Where featurename is the desired name, class is the feature class, eg
        'window', and size is a 2-tuple of (width,height).
        """
        raise NotImplementedError


    def _genFeatureSet (self):
        "Returns a newly generated feature set appropriate for this style"
        required = self._getFeatureList ()
        fsobj = self.getRegistry().getObject ('/featureset')
        fset = fsobj.getObject().generateContent (required, 
                                    self.getRegistry(), self.getPRand())
        return fset


    def getRegistry (self):
        "Returns the currently active registry"
        return self.registry

    def getPRand (self):
        "Returns the current PRand object"
        return self.prand

    def fillBuilding (self, building, registry, prand):
        "Fills the faces of a Building"
        
        ## HACK!
        walltype = lambda: random.choice ( [FRONTFACE, BACKFACE,
                                            MISCFACE, MISCFACE] )
        
        self._building = building

        self.registry = registry
        self.prand = prand
        
        featureset = self._genFeatureSet ()
        self._featureset = featureset

        obj = MetaObject.New ('building')
        
        # fill the walls
        for face in building.getFaceList ('walls'):
            obj.addObjectData (self.fillWall (face, walltype()))

        # fill the roof[s]
        for face in building.getFaceList ('roof'):
	    obj.addObjectData (featureset.roofFunc (face, registry, prand))

        obj.setTransform (building.transform)

        del (self._building)

        return obj
        

    def fillWall (self, face, type=MISCFACE):
        """Fills a wall-type Face with the given feature set. 
        Returns an Object list."""
        
        info = face.getWallInfo ()
        
        # generate the face
        size = (info.width, info.height)
        faceobj = self.fillRect (size, type)
        
        # transform the face back to its original location
        xform = Maths.Util.MakeXForm (info.origin, info.yrot)
        faceobj.applyTransformFunc (xform)

        return faceobj
                
        
	
    def fillRect (self, tr, type=MISCFACE):
	"Fills rect from O->tr with given feature set. Returns Object list"
	
	self.fill = Object.New ('fill')
        robj = self.chooseSocketValue ('MAT_WALLFACE', 
                                        self.getRegistry(), self.getPRand())
        self.fill.material = robj.getObject ()
	self.ncells,self.cellsize = _GetCellInfo (tr, self._featureset)
        self.rectsize = tr

	self.faceobject = MetaObject.New ('face')

	# call the user-defined stuff:
	self._startFace (type)
    
	for col in range (self.ncells[0]):
	    for row in range (self.ncells[1]):
		cellbase = Vec3(col*self.cellsize[0], row*self.cellsize[1], 0)
		self._fillCell (row, col, cellbase, type)

	self._endFace (type)
        
        del self.rectsize
        del self.cellsize
        del self.ncells

	self.faceobject.addObjectData (self.fill)
	return self.faceobject
    

    def getRectSize (self):
        """Returns the dimensions of the rectangle being filled.
        Valid between _startFace and _endFace."""
        return self.rectsize

    def getCellSize (self):
        """Returns the dimensions of the cells.
        Valid between _startFace and _endFace."""
        return self.cellsize

    def getCellCount (self):
        """Returns the number of cells in each direction.
        Valid between _startFace and _endFace."""
        return self.ncells
        
        
    def addFeature (self, featurename, cellbase):
	"Adds the named feature to the scene at the specified position"
        feature = self._featureset.getFeature (featurename)
	feature.object.translate (feature.getOffset (self.cellsize) + cellbase)
	self.faceobject.addObjectData (feature.getObject ())

	# fill in around the feature
	filltris = feature.genFill (cellbase, self.cellsize)
	for tri in filltris:
	    self.fill.addTri (tri[0], tri[1])


    def addObject (self, obj):
        "Adds the object data to the current face object"
        self.faceobject.addObjectData (obj)


    # overridable stuff
    def _startFace (self, facetype):
	"This is called once before the face is filled"
	pass
	
    def _fillCell (self, row, col, cellbase, facetype):
	"USER-DEFINED: should add objects with addFeature"
	raise NotImplementedError

    def _endFace (self, facetype):
	"This is called once after the face has been filled"
	pass


    
	
def _GetCellInfo (facesize, featureset):
    "Returns tuple: ( (ncellsw, ncellsh), (cellw, cellh) )"
    
    mincellsize = featureset.getMinCellSize()

    ncellsw = math.floor (facesize[0] / mincellsize[0])
    ncellsh = math.floor (facesize[1] / mincellsize[1])
    
    if ncellsw==0 or ncellsh==0:
        print 'WARNING: a face was too small to fit a feature on'
        if ncellsw==0: ncellsw = 1
        if ncellsh==0: ncellsh = 1

    cellw = facesize[0] / ncellsw
    cellh = facesize[1] / ncellsh

    return ( (int(ncellsw),int(ncellsh)), (cellw,cellh) )


