"""A 3D vector class with overloaded operators and useful functions
Targeted at robustness over speed, this does perform type checking where
it makes sense [eg. no divide by vector].

Operators:
 v1+v2,v1+=v1	addition
 v1-v2,v1-=v2	subtraction
 v1*s,v1*=s	scale
 v1/s,v2/=s	scale by 1/s
 -v1		negation
 abs(v1)	component-wise absolute value
 v1==v2		component-wise equality test [false if v2 is not a Vec3]
 v1!=v2		component-wise inequality test [true if v2 is not a Vec3]

Methods:
 v1.dot(v2)		dot product     [     v1 * v2    ]
 v1.cross(v2)		cross product   [     v1 x v2    ]
 v1.length()		magnitude of v1 [     | v1 |     ]
 v1.length_sq()		mag. squared    [   | v1 | ^ 2   ]
 v1.normalise()		normalise v1	[ v1 = v1 / |v1| ]
 v1.getnormalised()	returns the normalised vector (doesn't change v1)
 v1.scale(s)            scales the vector by s [a list of 3 items]

 v1.clone()		returns a new copy of v1
 v1.assign(v2)          copies v2 into v1

Pseudo-attributes:
 v1.x, v1.y and v1.z work as expected

Alex Mole, 2003
"""

import math
import Matrix


class Vec3:
    def __init__ (self, x=0., y=0., z=0.):
	# make sure these are floating point values
	self.x = float (x)
	self.y = float (y)
	self.z = float (z)

    def __repr__ (self):
	return 'Vec3(%f, %f, %f)' % (self.x, self.y, self.z)

    def __str__ (self):
	return '(%f, %f, %f)' % (self.x, self.y, self.z)

    def __hash__ (self):
        return hash ( (self.x, self.y, self.z) )

    def __eq__ (self, rhs):
	if not isVector (rhs):
	    return False
	return ((self.x == rhs.x) and (self.y == rhs.y) and (self.z == rhs.z))

    def __ne__ (self, rhs):
	if not isVector (rhs):
	    return True
	return ((self.x != rhs.x) or (self.y != rhs.y) or (self.z != rhs.z))

    # container
    def __len__ (self):
        return 3

    def __getitem__ (self, index):
        if index == 0:
            return self.x
        elif index == 1:
            return self.y
        elif index == 2:
            return self.z
        else:
            raise IndexError, 'vector index out of range'

    def __setitem__ (self, index, value):
        if index == 0:
            self.x = value
        elif index == 1:
            self.y = value
        elif index == 2:
            self.z = value
        else:
            raise IndexError, 'vector assignment index out of range'

    # vector addition
    def __add__ (self, rhs):
	if not isVector (rhs):
	    raise TypeError, "Attempt to add a non-vector to a vector"
	    
	return Vec3 (self.x+rhs.x, self.y+rhs.y, self.z+rhs.z)

    def __iadd__ (self, rhs):
	if not isVector (rhs):
	    raise TypeError, "Attempt to add a non-vector to a vector"

	self.x += rhs.x
	self.y += rhs.y
	self.z += rhs.z
	return self

    # vector subtraction
    def __sub__ (self, rhs):
	if not isVector (rhs):
	    raise TypeError, "Attempt to subtract a non-vector from a vector"

	return Vec3 (self.x-rhs.x, self.y-rhs.y, self.z-rhs.z)

    def __isub__ (self, rhs):
	if not isVector (rhs):
	    raise TypeError, "Attempt to subtract a non-vector from a vector"

	self.x -= rhs.x
	self.y -= rhs.y
	self.z -= rhs.z
	return self

    # scalar multiplication
    def __mul__ (self, rhs):
	# if it's a vec*mat, use Mat4 rmul
	if Matrix.isMatrix (rhs):
	    return rhs.__rmul__ (self)
	else:
	    s = float (rhs)		# make sure we're using a float
	    return Vec3 (self.x*s, self.y*s, self.z*s)

    def __imul__ (self, rhs):
	s = float (rhs)		# make sure we're using a float
	self.x *= s
	self.y *= s
	self.z *= s
	return self

    def __rmul__ (self, lhs):
	s = float (lhs)		# make sure we're using a float
	return 	Vec3 (self.x*s, self.y*s, self.z*s)
	
    # scalar division
    def __div__ (self, rhs):
	s = float (rhs)		# make sure we're using a float
	return self * (1.0 / s)
    
    def __idiv__ (self, rhs):	
	s = float (rhs)		# make sure we're using a float
	self *= 1.0 / s
	return self

    def dot (self, rhs):
	"Returns the dot product of self and rhs"
	if not isVector (rhs):
	    raise TypeError, "Vec3.dot called with a non-vector argument"
	return (self.x*rhs.x + self.y*rhs.y + self.z*rhs.z)

    def cross (self, rhs):
	"Returns the result of (_self_ x _rhs_)"
	if not isVector (rhs):
	    raise TypeError, "Vec3.cross called with a non-vector argument"
	return Vec3 (	self.y*rhs.z - self.z*rhs.y,
			self.z*rhs.x - self.x*rhs.z,
			self.x*rhs.y - self.y*rhs.x)

    def length (self):
	"Returns the modulus of the vector"
	return math.sqrt (self.x*self.x + self.y*self.y + self.z*self.z)

    def length_sq (self):
	"Returns the modulus of the vector, squared"
	return self.x*self.x + self.y*self.y + self.z*self.z

    def normalise (self):
	"Normalises the vector"
	self /= self.length()
	return self

    def getNormalised (self):
	"Returns a normalised copy of the vector"
	return self / self.length()

    def clone (self):
	"Returns a new copy of this vector"
	return Vec3 (self.x, self.y, self.z)

    def assign (self, other):
        "Copies other.[xyz] to self.[xyz]"
        self.x = other.x
        self.y = other.y
        self.z = other.z

    def scale (self, s):
        "Scales self by s, a list of 3 floats"
        assert len(s) == 3
        self.x *= s[0]
        self.y *= s[1]
        self.z *= s[2]
        return self
    
	
	
# type check
def isVector (v):
    return isinstance (v, Vec3)


# tries to return v as a Vec3 [v can be Vec3, list, tuple]
def MakeVector (v):
    if isVector (v):
	return v
    elif hasattr (v, '__len__') and len (v) == 3:
	return Vec3 (float(v[0]), float(v[1]), float(v[2]))


def GetMinMax (vlist):
    "Returns min,max component-wise of a list of Vec3"
    if not vlist:
        return Vec3(0,0,0),Vec3(0,0,0)
    
    max = vlist[0].clone ()
    min = vlist[0].clone ()
    for v in vlist:
        if v.x > max.x: max.x = v.x
        if v.y > max.y: max.y = v.y
        if v.z > max.z: max.z = v.z
        if v.x < min.x: min.x = v.x
        if v.y < min.y: min.y = v.y
        if v.z < min.z: min.z = v.z
    return min,max


def GetMin (vlist):
    "Returns min component-wise of a list of Vec3"
    if not vlist:
        return Vec3(0,0,0)
    
    min = vlist[0].clone ()
    for v in vlist:
        if v.x < min.x: min.x = v.x
        if v.y < min.y: min.y = v.y
        if v.z < min.z: min.z = v.z
    return min


def GetMax (vlist):
    "Returns max component-wise of a list of Vec3"
    if not vlist:
        return Vec3(0,0,0)
    
    max = vlist[0].clone ()
    for v in vlist:
        if v.x > max.x: max.x = v.x
        if v.y > max.y: max.y = v.y
        if v.z > max.z: max.z = v.z
    return max


