from PCD.Maths import Matrix,Vector3D
from PCD.Maths.Vector3D import Vec3
from PCD import Transform, Material
import copy


def New (name='Mesh'):
    "Creates and returns a new Object"
    return Object (name)


class Object(object):

    def __init__ (self, name='Mesh'):
    	self.name = name
	self.verts = []
	self.tris = []
        self.texfaces = []
        self.material = None
	self.transform = Transform.New()
	self._bb = (Vec3(0,0,0), Vec3(0,0,0))
        self._bb_valid = False

    
    def __repr__ (self):
	return 'Object(%s, %s, %s, %s)' % (self.name, 
                repr (self.verts), repr (self.tris),
                repr (self.texfaces))
    
    def __str__ (self):
	return '<Object: %s, %s, %s>' % (self.name, 
					    str (self.verts),
					    str (self.tris))

    
    def clone (self):
	"Returns a clone of this object"
	o = Object (self.name)

        o.transform.assign (self.transform)

        if self._bb_valid:
	    o._bb = (self._bb[0].clone(), self._bb[1].clone())
            o._bb_valid = True
        else:
            o._bb_valid = False
    
        o.verts = [v.clone() for v in self.verts]
        o.tris = copy.copy (self.tris)
        o.texfaces = copy.copy (self.texfaces)

        # not cloned!
        o.material = self.material
    
	return o

        
    def addVert (self, vert):
        "Returns index"
        self.verts.append (vert.clone ())
        self._bb_valid = False
        return len (self.verts) - 1
	
    def getVert (self, index):
	"Returns the numbered vertex"
        return self.verts[index]
	
        
    def addTri (self, tri, texface=None):
	"Adds a triangle [3 vertices] to the object"
	indices = [self.addVert (v) for v in tri[0:3]]
        self.addTriIndices (indices, texface)

    def addTriIndices (self, indices, texface=None):
	"Adds a triangle with supplied vertex indices"
        assert indices[0] < len (self.verts)
        assert indices[1] < len (self.verts)
        assert indices[2] < len (self.verts)
        
        # NOTE: texfaces get added first so that this test works
        if len (self.texfaces) == len (self.tris):
            if texface != None: 
                self.texfaces.append (tuple (texface[0:3]))
            elif self.material != None:
                self.texfaces.append ( ((0,0), (0,0), (0,0)) )
                
	self.tris.append (tuple (indices[0:3]))


    def addObjectData (self, other):
	"Adds the data from another object to this one"
	if other == None: return

	# need to remap triangles to new vertex list
        for i in xrange (len (other.tris)):
            tri = other.tris[i]
	    indices = []
	    for oldi in tri:
		newi = self.addVert (other.getVert (oldi))
		indices.append (newi)
    
            try: texface = other.texfaces[i]
            except IndexError: texface = None
            
	    self.addTriIndices (indices, texface)
            
        self._bb_valid = False


    def setTransform (self, t):
        "Assigns the Transform t to self"
        self.transform.assign (t)
        self._bb_valid = False
        
    def applyTransform (self):
	"Applies the transform to all the vertices and clears the transform"
        m = self.transform.getMatrix()
        self.verts = [v*m for v in self.verts]
        self.transform.clear ()
        self._bb_valid = False

    def applyScale (self):
        """Applies any scaling in the transform to the vertices, but leaves
        rotations and translations alone"""
        for v in self.verts:
            v.scale (self.transform.sca)
        self.transform.sca = Vec3(1,1,1)
        self._bb_valid = False

    def applyTransformFunc (self, transform):
	"Filters each vertex through _transform_ function"
	for vert in self.verts:
	    vert.assign (transform (vert))
        self._bb_valid = False

    def translate (self, t):
        "Translates the object by vector t"
        self.applyTransformFunc (lambda v,s=t: v+s)    
        self._bb_valid = False
    
    def calcBBox (self):
	"Calculates the bounding box for this object"
        if self._bb_valid:
            return
        self._bb = Vector3D.GetMinMax (self.verts)
        self._bb_valid = True
    
    def getBBox (self):
	"Returns the bounding box as tuple: (min, max)"
        if not self._bb_valid:
            self.calcBBox ()
	return self._bb

    def getHeight (self):
	"Returns height of the object (y-axis)"
        bb = self.getBBox ()            
	return bb[1].y - bb[0].y
	
    def getWidth (self):
	"Returns width of the object (x-axis)"
        bb = self.getBBox ()            
	return bb[1].x - bb[0].x
	

