"""Processes .pip object files

Alex Mole, 2003
"""

import sys
from PCD.Util import XMLParse
from PCD import Scene, Object, Material
from PCD.Maths import Vector3D


class ParseError(Exception):
    def __init__ (self, val):
        self.value = val
    def __str__ (self):
        return repr (self.val)
        

class PIPFileParser:
    "Parses a PIP geometry file and creates a scene"

    def __init__ (self):
        # the current list of vertices
        self.currvtxlist = []
        # the current Object
        self.currobject = None
        # the current Material
        self.currmaterial = None
        

    def load (self, filename):
        "Loads a .pip file into a Scene"
        # drop any previous scene info we had
        self.scene = Scene.New ()
        tree = XMLParse.XMLFileToTree (filename)

        try:
            self._parseScene (tree)
        except ParseError:
            print >>sys.stderr, 'ERROR: Parse error in PIP file '+filename
            raise
            
        return self.scene


    def getScene (self):
        "Returns the scene"
        return self.scene
    

    #### parse functions
        
    def _parseScene (self, tree):
        assert tree[0] == 'scene'

        # load all the textures
        for texlist in filter (lambda n: n[0]=='texlist', tree[3]):
            self._parseTexlist (texlist)

        # load all the materials
        for matlist in filter (lambda n: n[0]=='matlist', tree[3]):
            self._parseMatlist (matlist)
    
        # load all the objects
        for obj in filter (lambda n: n[0]=='object', tree[3]):
            self._parseObject (obj)

    
    def _parseTexlist (self, tree):
        assert tree[0] == 'texlist'
        for texture in tree[3]:
            self._parseTexture (texture)

    def _parseTexture (self, tree):
        assert tree[0] == 'texture'
        self.scene.newTexture (tree[1]['src'], tree[1]['name'])

    def _parseMatlist (self, tree):
        assert tree[0] == 'matlist'
        for material in tree[3]:
            self._parseMaterial (material)

    def _parseMaterial (self, tree):
        assert tree[0] == 'material'
        material = self.scene.newMaterial (tree[1]['name'])
        for texchannel in tree[3]:
            self._parseTexchannel (texchannel, material)
    
    def _parseTexchannel (self, tree, material):
        assert tree[0] == 'texchannel'
        channel = int (tree[1]['num'])
        tex = self.scene.getTexture (tree[1]['texture'])
        material.setTexture (channel, tex)

    def _parseObject (self, tree):
        assert tree[0] == 'object'
        obj = self.scene.newObject (tree[1]['name'])
        for child in tree[3]:
            if child[0] == 'transform': self._parseTransform (child, obj)
            elif child[0] == 'vtxlist': self._parseVtxlist (child, obj)
            elif child[0] == 'trilist': self._parseTrilist (child, obj)
            elif child[0] == 'matref': self._parseMatref (child, obj)
            else: 
                raise ParseError, "Found invalid tag '"+child[0]+"'"
                
        obj.applyScale () 
        obj.calcBBox ()

    def _parseTransform (self, tree, obj):
        assert tree[0] == 'transform'
        for child in tree[3]:
            if child[0] == 'position': 
                self._parsePosition (child, obj.transform)
            elif child[0] == 'eulerrot':
                self._parseEulerrot (child, obj.transform)
            elif child[0] == 'scale':
                self._parseScale (child, obj.transform)
            else: 
                raise ParseError, "Found invalid tag '"+child[0]+"'"

    def _parsePosition (self, tree, transform):
        assert tree[0] == 'position'
        transform.setPosition ( [float(s) for s in tree[2].split()] )

    def _parseEulerrot (self, tree, transform):
        assert tree[0] == 'eulerrot'
        transform.setRotation ( [float(s) for s in tree[2].split()] )

    def _parseScale (self, tree, transform):
        assert tree[0] == 'scale'
        transform.setScale ( [float(s) for s in tree[2].split()] )

    def _parseVtxlist (self, tree, obj):
        assert tree[0] == 'vtxlist'
        for vtx in tree[3]:
            self._parseVtx (vtx, obj)

    def _parseVtx (self, tree, obj):
        assert tree[0] == 'vtx'
        x,y,z = [float(s) for s in tree[2].split()]
        obj.verts.append (Vector3D.Vec3 (x, y, z))

    def _parseTrilist (self, tree, obj):
        assert tree[0] == 'trilist'
        texinfo = bool (int (tree[1].get ('texinfo', '0')))
        for tri in tree[3]:
            self._parseTri (tri, obj, texinfo)

    def _parseTri (self, tree, obj, texinfo):
        assert tree[0] == 'tri'
        data = tree[2].split()
        indices = tuple ([int(i) for i in data[0:3]])
        if texinfo:
            texface = [ tuple (map (lambda v:float(v), s.split(',')))
                                                        for s in data [3:6] ]
        else:
            texface = None
        obj.addTriIndices (indices, texface)

    def _parseMatref (self, tree, obj):
        assert tree[0] == 'matref'
        material = self.scene.getMaterial (tree[1]['material'])
        obj.material = material
        



class PIPFileWriter:
    "Saves a Scene in .pip file format"

    indentsize = 4      #xml indent size

    def __init__ (self):
        self.outfile = sys.stdout       # default: write to screen
        self.indentlevel = 0

    def indent (self):
        self.indentlevel += PIPFileWriter.indentsize

    def unindent (self):
        self.indentlevel -= PIPFileWriter.indentsize
        if self.indentlevel < 0:
            self.indentlevel = 0

    def write (self, s):
        i = ' ' * self.indentlevel
        self.outfile.write (i + s + '\n')

    def write_i (self, s):
        self.write (s)
        self.indent ()

    def u_write (self, s):
        self.unindent ()
        self.write (s)

    def writeList (self, type, items, fun, attribs=None):
        "writes a general list, calling fun for each item"
        if attribs == None:
            self.write_i ('<%slist>' % type)
        else:
            self.write_i ('<%slist %s>' % (type, attribs))
            
        for i in items: fun (i)
        
        self.u_write ('</%slist>' % type)

    
    def writeVertex (self, v):
        self.write ('<vtx>%f %f %f</vtx>' % (v.x, v.y, v.z) )

    def writeTri (self, tri, texface=None):
        text = '%d %d %d' % tri
        if texface:
            uvs = tuple ([s for uv in texface for s in uv]) # flatten it :)
            text += ' %f,%f %f,%f %f,%f' % uvs
        self.write ('<tri>%s</tri>' % text)
        
    def writeTriList (self, object):
        "writes the <trilist>"
        if self.writetexinfo: self.write_i ('<trilist texinfo="1">')
        else: self.write_i ('<trilist>')
        
        for i in range (len (object.tris)):
            if self.writetexinfo: texface = object.texfaces[i]
            else: texface = None
            self.writeTri (object.tris[i], texface)
        
        self.u_write ('</trilist>')

        
    def writeTexture (self, tex):
        self.write ('<texture name="%s" src="%s" />' % (tex.name,tex.filename))
        
    def writeMaterial (self, mat):
        self.write_i ('<material name="%s">' % mat.getSaveName())
        textures = mat.getTextures ()
        for i in range (len (textures)):
            t = textures[i]
            if t: self.write ('<texchannel num="%d" texture="%s" />' %
                                                                (i, t.name) )
        self.u_write ('</material>')

    def writeTransform (self, t):
        self.write_i ('<transform>')
        self.write ('<position>%f %f %f</position>' % 
                                                (t.pos.x, t.pos.y, t.pos.z))
        self.write ('<eulerrot>%f %f %f</eulerrot>' % 
                                                (t.rot.x, t.rot.y, t.rot.z))
        self.write ('<scale>%f %f %f</scale>' % 
                                                (t.sca.x, t.sca.y, t.sca.z))
        self.u_write ('</transform>')
        
    def save (self, scene, filename=None):
        "Saves the scene. If filename is None, the output is sent to STDOUT"
        
        if filename != None:
            self.outfile = file (filename, 'w')
            
        # write the xml header
        self.write ('<?xml version="1.0"?>')

        # scene object is top-level object
        self.write_i ('<scene>')

        # write the texture list
        self.writeList ('tex', scene.getTextures (), self.writeTexture)

        # write the material list
        self.writeList ('mat', scene.getMaterials (), self.writeMaterial)

        # get the list of objects in the scene
        objects = scene.getObjects()

        for obj in objects:
            self.write_i ('<object name="%s">' % obj.name)
            self.writeTransform (obj.transform)

            # write the vertex list
            self.writeList ('vtx', obj.verts, self.writeVertex)

            # write the triangle list
            if obj.texfaces:
                if len(obj.texfaces) != len(obj.tris):
                    print 'WARNING: tris:%d, texfaces:%d' % (
                                    len(obj.tris), len(obj.texfaces))

                self.writetexinfo = True
            else:
                self.writetexinfo = False

            self.writeTriList (obj)
        
            if obj.material:
                self.write ('<matref material="%s" />' % 
                                                obj.material.getSaveName())

            self.u_write ('</object>')

        self.u_write ('</scene>')

        if filename != None:
            self.outfile.close()
            self.outfile = sys.stdout
        


def Load (filename):
    "Loads a .pip file and returns a Scene"
    p = PIPFileParser()
    p.load (filename)
    return p.getScene ()
    

def Save (scene, filename):
    "Saves a Scene to a .pip format file"
    p = PIPFileWriter()
    p.save (scene, filename)
        

    
# testing....
if __name__ == '__main__':
    if len(sys.argv) > 1: scenename = sys.argv[1]
    else: scenename = '../data/foo.pip'

    print 'Loading scene from', scenename
    scene = Load (scenename)
    print scene
    Save (scene, '../data/foo_save.pip')




