#
# The registry structure
#
# This is basically a mini-filesystem complete with links and "properties".
# It is designed to look like a modern "unixy" FS such as ReiserFS, so objects
# are accessed with paths like: '/feature/window/decorative/stained-glass',
# and their metadata like '/path/to/object:meta-name'
# 

import random



def New ():
    "Returns a new Registry object"
    return Registry ()


class PathError (Exception):
    def __init__ (self, s):
        Exception.__init__ (self)
        self.s = s
    def __str__ (self):
        return repr (self.s)


class MetaData (object):
    "Represents a single item of metadata"

    def __init__ (self, name, type):
        self.name = name
        self.type = type
        self.value = None



class RegObject (object):
    """An object stored in a Registry.
    If the registered object has a method called getRegMeta(), this is called
    and should return a list of tuples: ('metadata_name','type')"""

    def __init__ (self, name, obj):
        self.name = name
        self.obj = obj
        self._meta = {}
        try:
            allmeta = obj.getRegMeta()
            for meta in allmeta:
                self._meta[meta[0]] = MetaData (meta[0], meta[1])
        except (AttributeError,TypeError):
            # obj isn't interested in metadata
            pass

    def __str__ (self):
        return repr (self.obj)

    def __repr__ (self):
        return 'RegObject(',self.name,',',repr(self.obj)+')'

    def getObject (self):
        return self.obj

    def getMeta (self, name):
        try:
            return self._meta[name]
        except KeyError:
            raise PathError, 'Metadata does not exist: ' + name
        

        
class RegDir (object):

    def __init__ (self, name):
        self.name = name
        self._subdirs = {}
        self._objects = {}


    def _addSubDir (self, name):
        "Adds a subdir to this with the given name"
        if name in self._subdirs:
            raise PathError, "Cannot create dir: dir already exists: "+name
        if name in self._objects:
            raise PathError, "Cannot create dir: object already exists: "+name
            
        self._subdirs[name] = RegDir (name)
        return self._subdirs[name]
        

    def _addObject (self, name, obj):
        "Adds the object to this dir with the supplied name"
        if name in self._subdirs:
            raise PathError, "Cannot create object: dir already exists"
        if name in self._objects:
            raise PathError, "Cannot create object: object already exists"
            
        self._objects[name] = RegObject (name, obj)
        return self._objects[name]
        
        
    def createPath (self, pathlist):
        """Ensures that the specified path exists, creating dirs as necessary.
        Returns the bottom-level folder"""
        if len (pathlist) == 0:
            raise PathError, "Empty path"
        
        if pathlist[0] not in self._subdirs:
            self._addSubDir (pathlist[0])

        if len (pathlist) > 1:
            return self._subdirs[pathlist[0]].createPath (pathlist[1:])
        else:
            return self._subdirs[pathlist[0]]


    def createObject (self, pathlist, obj):
        "Adds the object at the specified path and returns the new RegObject."
        if len (pathlist) == 0:
            raise PathError, "Empty path"

        parent = self
        if len (pathlist) > 1:
            parent = self.createPath (pathlist[:-1])
        
        return parent._addObject (pathlist[-1], obj)


    def getDir (self, pathlist):
        """Returns the named directory"""
        if len (pathlist) == 0:
            raise PathError, "Empty path"
        
        if len (pathlist) == 1:
            
            if pathlist == ['']:
                return self
                
            try:
                return self._subdirs[pathlist[0]]
            except KeyError:
                raise PathError, "Dir does not exist: "+pathlist[0]
        
        try:
            return self._subdirs[pathlist[0]].getDir (pathlist[1:])
        except KeyError:
            raise PathError, "Dir does not exist: " + pathlist[0]


    def getObject (self, pathlist):
        """Returns the named object [or metadata].
        pathlist should be a list of path items, as returned by SplitPath"""
        if len (pathlist) == 0:
            raise PathError, "No path specified"
        
        if len (pathlist) == 1:
            try:
                name,meta = SplitName (pathlist[0])
                if meta == None:
                    return self._objects[name]
                else:
                    return self._objects[name].getMeta (meta)
                    
            except KeyError:
                raise PathError, "Object does not exist: "+name
        
        try:
            return self._subdirs[pathlist[0]].getObject (pathlist[1:])
        except KeyError:
            raise PathError, "Dir does not exist: " + pathlist[0]
            

    def list (self, pathlist):
        """Returns the alphabetised 'directory listing' for the specified path.
        pathlist should be a list of path items, as returned by SplitPath.
        Returns tuple like (subdirs,objects)."""

        if len (pathlist) == 0 or pathlist==['']:
            dirs = self._subdirs.keys()
            dirs.sort ()
            objects = self._objects.keys()
            objects.sort ()
            return (tuple(dirs), tuple(objects))

        if pathlist[0] not in self._subdirs:
            raise PathError, "Dir does not exist: " + pathlist[0]

        return self._subdirs[pathlist[0]].list (pathlist[1:])


    def walk (self):
        """Returns a list of all the objects under this dir, with paths
        relative to this."""
        objs = self._objects.keys()
        objs.sort ()
        
        dirs = self._subdirs.keys()
        dirs.sort ()
        for d in dirs:
            subobjs = self._subdirs[d].walk()
            subobjs = [JoinPath (d,s) for s in subobjs]
            objs.extend (subobjs)

        return objs


    def dirExists (self, pathlist):
        "Returns true if the named dir exists."
        if len (pathlist) == 0:
            raise PathError, "Empty path"

        if len (pathlist) == 1:
            return pathlist[0] in self._subdirs

        try:
            return self._subdirs[pathlist[0]].dirExists (pathlist[1:])
        except KeyError:
            return False


    def objectExists (self, pathlist):
        "Returns true if the named object exists."
        if len (pathlist) == 0:
            raise PathError, "Empty path"

        if len (pathlist) == 1:
            name,meta = SplitName (pathlist[0])
            return name in self._objects
            
        try:
            return self._subdirs[pathlist[0]].objectExists (pathlist[1:])
        except KeyError:
            return False

    
    def pathExists (self, pathlist):
        "Returns true if something [dir or object] exists at that path"
        return self.dirExists (pathlist) or self.objectExists (pathlist)



#############################################################################
# Registry class                                                            #
#############################################################################

class Registry (object):

    def __init__ (self):
        self._root = RegDir ('')


    def registerObject (self, obj, path=None):
        """Registers obj to the specified path. If no path has been specified,
        obj.getRegName() is used to obtain it."""
        try:
            if path == None:
                try:
                    path = obj.getRegName ()
                except (AttributeError, TypeError):
                    raise PathError, "Unable to get path"
                    
            plist = SplitPath (path)
            regobj = self._root.createObject (plist, obj)
            
            return regobj
            
        except PathError, e:
            raise PathError, "Cannot register object: "+str(e)


    def getObject (self, path):
        "Returns the named object or metadata"
        plist = SplitPath (path)
        try:
            return self._root.getObject (plist)
        except PathError, e:
            raise PathError, "Cannot get object '"+path+"': '"+str(e)


    def listDir (self, path):
        "Returns the names in a directory in a tuple: (folders,objects)"
        plist = SplitPath (path)
        return self._root.list (plist)


    def walkDir (self, path):
        """Returns the names of the objects in all subdirectories of this path,
        relative to the path."""
        plist = SplitPath (path)
        topdir = self._root.getDir (plist)
        return topdir.walk ()


    def chooseObject (self, path, prand=None, reqname=''):
        """Returns an object at random from the supplied path.
        Uses the supplied PRand object with the request name [or Random]."""
        objs = self.walkDir (path)
        if prand is not None:
            i = prand.num (reqname, len (objs))
            fullpath = JoinPath (path, objs[i])
        else:
            fullpath = JoinPath (path, random.choice(objs))
        return self.getObject (fullpath)
        
        




##############################################################################
#  Helper functions                                                          #
##############################################################################

def SplitPath (path):
    "Returns a path '/a/b/c:d' as ['a','b','c:d']"
    if len(path) == 0:
        raise PathError, 'Empty path'

    # metadata names may contain '/'
    if path.find (':') != -1:
        path,meta = SplitName (path)
        p = path.split ('/')
        p[-1] = p[-1] + ':' + meta
    else:
        p = path.split ('/')

    # a leading / will introduce an extra '' path element
    if path[0] == '/':
        p = p[1:]
    
    return p


def SplitName (name):
    "Returns a list: [filename, metadataname], where metadataname can be None"
    if len (name) == 0:
        raise PathError, 'Zero length path'
        
    l = name.split (':', 1)

    if len (l[0]) == 0 or l[0][-1] == '/':
        raise PathError, 'Name missing from path: '+name

    if len (l) == 1:
        l.append (None)
    return l


def DirName (path):
    """Returns the path to an object: '/a/b/c:d' -> '/a/b/'.
    Note that while '/a/b/' will return itself, '/a/b' will return just '/a/'.
    Take care to avoid this situation."""
    # special cases
    if path == '':
        raise PathError, 'Empty path'
    if path == '/':
        return path
    if path.find ('/') == -1:
        return ''

    s = reduce (lambda a,b: a+'/'+b, SplitPath (path)[:-1])
    if len (s) > 0: 
        s = s + '/'
    s = '/' + s
    return s


def ObjectName (path):
    "Returns the name of an object: '/a/b/c:d' -> 'c'."
    nm = SplitPath (path)[-1]
    if len (nm) == 0:
        raise PathError, 'Missing filename in path: ' + path
    n = SplitName (nm)[0]
    if len (n) == 0:
        raise PathError, 'Missing filename in path: ' + path
    return n
    
    
def MetaName (path):
    """Returns the name of the metadata in a path: '/a/b/c:d' -> 'd'.
    Returns None if there was no metadata specified."""
    nm = SplitPath (path)[-1]
    if len (nm) == 0:
        raise PathError, 'Missing filename in path: ' + path
    m = SplitName (nm)[1]
    return m
    

def JoinPath (a, b, *args):
    """Works much the same as os.path.join().
    Note that if you pass crazy things like '////' they won't be caught."""
    path = a
    lumps = [b] + list(args)
    for lump in lumps:
        if path[-1] == '/':
            if lump[0] == '/':
                path = path + lump[1:]
            else:
                path = path + lump
        else:
            if lump[0] == '/':
                path = path + lump
            else:
                path = path + '/' + lump
    return path
    



