"""
Module responsible for analysing a "Scene" and converting it into a tree
of buildings and their faces.

Alex Mole, 2004
"""

import math
from PCD.Maths.Vector3D import Vec3
from PCD import Transform
import PCD.Maths.Util



class WallInfo:
    "A class to analyse and report info about a wall-type Face"

    def __init__ (self, face):
        
        # v0.1: just handle y-axis aligned rectangles
        
        if (math.fabs (face.normal.y) > 0.0001):
            raise ValueError, "Face too complex to fill (non xz normal)"
        if (len (face.verts) != 4):
            raise ValueError, "Face too complex to fill (%d verts)" % \
                                                            len (face.verts)
        # we need to work out the most appropriate vertex to use as the 
        # bottom-left. 
        # face.verts will be wound anticlockwise, so we can infer the bl's
        # position based on any three vertices.
        a,b,c,d = face.verts
        eq = lambda a,b: math.fabs(a-b) < 0.0001
        if eq (a.y, b.y):
            assert eq (d.y, c.y)
            assert eq (a.x,d.x) and eq (a.z,d.z)
            assert eq (b.x,c.x) and eq (b.z,c.z)
            if a.y > d.y:
                bl,tr = c,a
            else:
                bl,tr = a,c
                
        else:
            assert eq (a.y,d.y) and eq (b.y,c.y)
            assert eq (a.x,b.x) and eq (a.z,b.z)
            assert eq (c.x,d.x) and eq (c.z,d.z)
            if a.y > b.y:
                bl,tr = b,d
            else:
                bl,tr = d,b

        # now we can work out the dimensions of the face
        diag = tr - bl
        self.width = math.sqrt (diag.x**2 + diag.z**2)
        self.height = diag.y
        
        self.origin = bl

        self.yrot = math.atan2 (face.normal.x, face.normal.z)
        

        

class Face:
    
    def __init__ (self, verts):
        self.verts = [v.clone() for v in verts]
        self.normal = PCD.Maths.Util.FaceNormal (verts)
        

    def __str__ (self):
        return '(' + str(self.normal) + ', ' + repr(self.verts) + ')'


    def tryAddTri (self, tri):
        """Tries to add the triangle to the face. 
        Returns False if the face could not be added (if it's not part of
                the face)."""

        # two of triangle's verts should match two consecutive face verts
        # find the first one
        for i,v in enumerate (self.verts):
            try:
                j = tri.index (v)
                break;
            except ValueError:
                pass
        else:
            # no matched verts
            return False

        # range checked vert accessors
        vself = lambda i: self.verts[i % len (self.verts)]
        vtri = lambda i: tri[i % len (tri)]
        
        # self[i] is the same as tri[j]
        # therefore, self[i+1] must match tri[j-1] for them to be attached
        # (because of consistent windings)
        if vself (i+1) == vtri (j-1):
            insert_before = (i+1) % len (self.verts)
            new_vtx = vtri (j+1)
        elif vself (i-1) == vtri (j+1):
            insert_before = i
            new_vtx = vtri (j-1)
        else:
            return False

        # the triangle is attached. now to test the normal.
        # testing for equality would be inaccurate due to floating point
        # errors. instead we test a threshold difference.
        normal = PCD.Maths.Util.FaceNormal (tri)
        d = (normal - self.normal).length_sq ()
        if d > 0.001:
            return False
    
        # we have a valid triangle. tri[j+1] is the new vertex.
        self.verts.insert (insert_before, new_vtx.clone ())
        
        return True


    def addAllTrisInList (self, tris):
        "Add all cofacial tris, removing them from the original list"

        while True:
            for i,t in enumerate (tris):
                if self.tryAddTri (t):
                    tris.pop (i)
                    break
            else:
                # no more cofacial triangles
                return

    
    def optimise (self):
        "Removes unnecessary vertices"

        # remove b where abc are colinear
        while True:
            for i,v0 in enumerate (self.verts):
                v1 = self.verts[(i+1)%len(self.verts)]
                v2 = self.verts[(i+2)%len(self.verts)]

                e0 = v1 - v0
                e1 = v2 - v0

                nz = lambda f: math.fabs(f) > 0.0001
                
                # if we have pairs of zero&nonzero then this isn't valid
                f = lambda a,b: ( (nz(a) and not nz(b)) or
                                  (nz(b) and not nz(a)) )
                l = [1 for a,b in zip(e0,e1) if f(a,b)]
                if len (l) != 0:
                    continue
                    
                # get a list of the component-wise ratios
                r = [a/b for a,b in zip(e0,e1) if (nz(a) and nz(b))]

                # if there are no non-zero ratios, the verts are co-incident,
                # and shouldn't ever have been added to the face....
                if len(r) > 0:
                    # v1 is not linearly between v0 and v2
                    if r[0] < 0 or r[0] > 1:
                        continue

                    def listeq (l):
                        for i in range(len(l)-1):
                            if math.fabs (l[i] - l[i+1]) > 0.0001:
                                return False
                        return True
                        
                    # are all the ratios equal?
                    if not listeq (r):
                        continue

                # v1 is unnecessary
                self.verts.remove (v1)

                # restart the operation so we have a safe state....
                break
                    
            else:
                # we're finished
                break
        

    def getWallInfo (self):
        "Returns a WallInfo structure. Only works for wall-type faces"
        return WallInfo (self)


    def getEdges (self):
        "Returns a ist of edge vectors: [v1-v0, v2-v1, ..., v0-vN]"
        v2 = self.verts*2   # double the list so that we loop round easy
        return [ (v2[i+1] - v2[i]) for i in range (len (self.verts)) ]


                
        

class Building (object):

    def __init__ (self):
        self.faces = {}
        self.faces['roof'] = []
        self.faces['walls'] = []
        self.faces['ground'] = []     # for the ground face- normally ignored
        self.transform = Transform.New ()
        
        
    def addFace (self, face):
        "Chooses which of the lists to add the face to based on its normal"
        
        if face.normal.y > 0.3:
            # pointing "up-ish", so it's a roof
            self.faces['roof'].append (face)
        elif math.fabs (face.normal.y) <= 0.3:
            # pointing "sideways-ish", so it's a wall
            self.faces['walls'].append (face)
        else:
            # pointing "down-ish", so it's a ground poly (normally ignored)
            self.faces['ground'].append (face)


    def getFaceList (self, listname):
        "Returns the contents of the named list [can be an empty list]"
        try:
            return self.faces[listname]
        except KeyError:
            return []

    def setTransform (self, t):
        self.transform.assign (t)
            


def Generate (scene):
    "Returns the scene tree for the supplied Scene object"

    # version 0.1: tree has 3 levels- root, buildings, faces
    #              each input object is considered to be a building
    
    scenetree = []
    
    objects = scene.getObjects ()

    for obj in objects:
        building = _GenerateBuildingTree (obj)
        if building != None:
            scenetree.append (building)

    return scenetree


def _GenerateBuildingTree (object):
    "Returns a Building object based on the supplied Object"
    
    # make a list of the actual triangles (not just index lists)
    tris = []
    for itri in object.tris:
       tris.append ([object.getVert (itri[0]),
                     object.getVert (itri[1]),
                     object.getVert (itri[2])])

    # if the object is empty, return None
    if len(tris) == 0:
        return None

    building = Building ()

    building.setTransform (object.transform)
    
    # we now need to group the triangles into faces
    while len (tris) > 0:
        f = Face (tris.pop (0))
        f.addAllTrisInList (tris)
        f.optimise ()
        building.addFace (f) 
      
    return building


