from PCD.Registry import *
from PCD import PRand
import unittest


class RegistryTestCase (unittest.TestCase):

    def testSplitPath (self):
        self.assertRaises (PathError, SplitPath, '')
        self.assertEqual (SplitPath ('/'), [''])
        self.assertEqual (SplitPath ('/a'), ['a'])
        self.assertEqual (SplitPath ('/a/'), ['a', ''])
        self.assertEqual (SplitPath ('/a/b/'), ['a', 'b', ''])
        self.assertEqual (SplitPath ('/a/b/c'), ['a', 'b', 'c'])
        self.assertEqual (SplitPath ('/a/b/c:d'), ['a', 'b', 'c:d'])
        self.assertEqual (SplitPath ('/a/b/c:d/e'), ['a', 'b', 'c:d/e'])

    def testSplitName (self):
        self.assertRaises (PathError, SplitName, '')
        self.assertRaises (PathError, SplitName, 'a/')
        self.assertRaises (PathError, SplitName, 'a/:bla')
        self.assertRaises (PathError, SplitName, ':bla')
        self.assertEqual (SplitName ('a'), ['a', None])
        self.assertEqual (SplitName ('a/b'), ['a/b', None])
        self.assertEqual (SplitName ('/a/b'), ['/a/b', None])
        self.assertEqual (SplitName ('a:bla'), ['a', 'bla'])
        self.assertEqual (SplitName ('/a/b:bla'), ['/a/b', 'bla'])
        self.assertEqual (SplitName ('/a/b:bla/foo'), ['/a/b', 'bla/foo'])

    def testDirName (self):
        self.assertRaises (PathError, DirName, '')
        self.assertEqual (DirName ('/a/b/c:d'), '/a/b/')
        self.assertEqual (DirName ('/a/b/'), '/a/b/')
        self.assertEqual (DirName ('/'), '/')
        self.assertEqual (DirName ('a'), '')

    def testObjectName (self):
        self.assertRaises (PathError, ObjectName, '')
        self.assertRaises (PathError, ObjectName, '/')
        self.assertRaises (PathError, ObjectName, '/a/b/')
        self.assertRaises (PathError, ObjectName, '/a/b/:bla')
        self.assertEqual (ObjectName ('a'), 'a')
        self.assertEqual (ObjectName ('/a'), 'a')
        self.assertEqual (ObjectName ('/a/b/c'), 'c')
        self.assertEqual (ObjectName ('/a/b/c:d'), 'c')
        self.assertEqual (ObjectName ('/a/b/c:d/e/f'), 'c')

    def testMetaName (self):
        self.assertRaises (PathError, MetaName, '')
        self.assertRaises (PathError, MetaName, '/')
        self.assertRaises (PathError, MetaName, '/a/b/')
        self.assertRaises (PathError, MetaName, '/a/b/:bla')
        self.assertEqual (MetaName ('a'), None)
        self.assertEqual (MetaName ('/a'), None)
        self.assertEqual (MetaName ('/a/b/c'), None)
        self.assertEqual (MetaName ('/a/b/c:'), '')
        self.assertEqual (MetaName ('/a/b/c:d'), 'd')
        self.assertEqual (MetaName ('/a/b/c:d/e/f'), 'd/e/f')

    def testJoinPath (self):
        self.assertEqual (JoinPath ('a', 'b'), 'a/b')
        self.assertEqual (JoinPath ('/', 'a'), '/a')
        self.assertEqual (JoinPath ('a', 'b', 'c'), 'a/b/c')
        self.assertEqual (JoinPath ('/a', 'b', 'c'), '/a/b/c')
        self.assertEqual (JoinPath ('/a/b', 'c'), '/a/b/c')
        self.assertEqual (JoinPath ('/a/b/', 'c'), '/a/b/c')
        self.assertEqual (JoinPath ('/a/b/', '/c'), '/a/b/c')
        self.assertEqual (JoinPath ('/a/b/', 'c:d', 'e'), '/a/b/c:d/e')


    #### Registry class tests

    def testRegisterObject (self):
        r = Registry ()
        o = 'I AM A TEST OBJECT'
        ro = r.registerObject (o, '/features/fridge')
        assert ro.getObject() is o
        self.assertRaises (PathError, r.registerObject, 
                                                'foo', '/features/fridge')
        
    def testGetObject (self):
        r = Registry ()
        o = 'I AM A TEST OBJECT'
        r.registerObject (o, '/features/fridge')
        go = r.getObject ('/features/fridge')
        assert go.getObject() is o
        self.assertRaises (PathError, r.getObject, '/grapefruit')

    def testListDir (self):
        r = Registry ()
        r.registerObject ('foo0', '/features/fridge')
        r.registerObject ('foo1', '/features/turnip')
        r.registerObject ('foo2', '/features/grapefruit')
        r.registerObject ('foo3', '/features/cdrom')
        r.registerObject ('foo4', '/features/aeroplane')
        r.registerObject ('foo5', '/features/doors/fancy')
        r.registerObject ('foo6', '/features/doors/plain')
        r.registerObject ('foo7', '/features/windows/porthole')
        r.registerObject ('foo8', '/curtain')
        dirs,objs = r.listDir ('/features')
        self.assertEqual (dirs, ('doors', 'windows'))
        self.assertEqual (objs, 
                    ('aeroplane', 'cdrom', 'fridge', 'grapefruit', 'turnip'))
        dirs,objs = r.listDir ('/')
        self.assertEqual (dirs, ('features',))
        self.assertEqual (objs, ('curtain',))
        self.assertRaises (PathError, r.listDir, '')
        self.assertRaises (PathError, r.listDir, '/curtain')
        self.assertRaises (PathError, r.listDir, '/aardvark')

    def testChooseObject (self):
        r = Registry ()
        objs = 'foo','bar','baz','chipmunk'
        dirname = '/features/stuff'
        for obj in objs:
            r.registerObject (obj, JoinPath(dirname,obj))
        pr = PRand.New ()
        o = r.chooseObject (dirname, pr, 'req0')
        assert o.getObject() in objs
        assert o.getObject() is r.chooseObject (dirname, pr,'req0').getObject()



def suite ():
    return unittest.makeSuite (RegistryTestCase)
    

if __name__ == "__main__":
    unittest.main ()

    
