from PCD import MetaObject, Object
from PCD.Architecture import Roof
from PCD.Maths.Vector3D import Vec3
import math, operator


class SimpleApex (Roof.Func):
    "Simple apex roof"

    def __init__ (self):
        self._regpath = '/roofs'
        self._regname = 'apex-simple'
        self._regsockets = [ ('mat', '/materials/roof/tiled'),
                             ('MAT_WALLFACE', '/materials/wall', True) ]

    def _isRect (self, face):
        "Returns true if face is rectangular"
        # v0.1: just check vert count ;)
        return len (face.verts) == 4

    def _genTexTri (self, obj, v):
        """Adds a textured tri to obj. Texcoords are calculated based on 
        worldspace. The first vert is the bottom-left of the texture."""
        # calc texcoords
        t = [ (0,0) ]   # first is always 0,0
        t.append ( ((v[1]-v[0]).length(), 0) )
        # top texcoord is tricky- work out the angle using cosine rule
        e = [v[1]-v[0], v[2]-v[1], v[0]-v[2]]
        costheta = ((e[0].length_sq() + e[2].length_sq() - e[2].length_sq()) /
                    (2 * e[0].length() * e[2].length()))
        x = e[2].length() * costheta
        y = e[2].length() * math.sin (math.acos (costheta))        
        t.append ( (x, y) )
        obj.addTri ( (v[0], v[1], v[2]), (t[0], t[1], t[2]) )
        
    def _genTexQuad (self, obj, v):
        """Adds a textured quad to obj. Texcoords are calculated based on 
        worldspace. The first vert is the bottom-left of the texture."""
        # calc texcoords
        t = [ (0,0) ]   # first is always 0,0
        t.append ( ((v[1]-v[0]).length(), 0) )
        t.append ( ((v[1]-v[0]).length(), (v[2]-v[1]).length()) )
        t.append ( (0, (v[2]-v[1]).length()) )
        obj.addTri ( (v[0], v[1], v[2]), (t[0], t[1], t[2]) )
        obj.addTri ( (v[0], v[2], v[3]), (t[0], t[2], t[3]) )
        

    def _generateRect (self, face, registry, prand):
        "Generate a roof for rectangular cross section buildings"
        assert self._isRect (face)
        
        # rearrange verts so that v0-v1 is a short edge
        edges = face.getEdges ()
        lengths_sq = [v.length_sq() for v in edges]
        if (lengths_sq[0] + lengths_sq[2]) > (lengths_sq[1] + lengths_sq[3]):
            verts = face.verts[1:] + face.verts[:1]
            edges = edges[1:] + edges[:1]
        else:
            verts = face.verts

        # work out the apex points
        height = prand.val ('roof-apex-height', 0.2, 0.6) * edges[0].length()
        a0 = verts[0] + (edges[0] * 0.5)
        a0.y += height
        a1 = verts[2] + (edges[2] * 0.5)
        a1.y += height
        
        # add the sides
        sides = Object.New ('roof')
        robj = self.chooseSocketValue ('mat', registry, prand)
        sides.material = robj.getObject ()
        self._genTexQuad (sides, (verts[1], verts[2], a1, a0))
        self._genTexQuad (sides, (verts[3], verts[0], a0, a1))
        
        # add the ends
        ends = Object.New ('ends')
        robj = self.chooseSocketValue ('MAT_WALLFACE', registry, prand)
        ends.material = robj.getObject ()
        self._genTexTri (ends, (verts[0], verts[1], a0))
        self._genTexTri (ends, (verts[2], verts[3], a1))
	
        o = MetaObject.New ('roof')
        o.addObjectData (sides)
        o.addObjectData (ends)
        return o
        
    def _generatePoly (self, face, registry, prand):
        "Generate a roof for buildings with arbitrary cross-sections"
        o = Object.New ('roof')
        robj = self.chooseSocketValue ('mat', registry, prand)
        o.material = robj.getObject ()
        
        verts = face.verts
        centroid = reduce (operator.add, verts) / float(len (verts))

        # raise the centroid
        centroid.y += prand.val ('roof-peak-height', 1, 4)
        
        for i in range (len (verts)):
            ni = (i+1) % len(verts)
            self._genTexTri (o, (verts[i], verts[ni], centroid))
	
        return o
        

    def generateContent (self, face, registry, prand):
        
        if self._isRect (face):
            return self._generateRect (face, registry, prand)
        else:
            return self._generatePoly (face, registry, prand)



class SlopedApex (SimpleApex):
    "Apex roof with all four sides sloping"

    def __init__ (self):
        SimpleApex.__init__ (self)
        self._regname = 'apex-sloped'

    def _generateRect (self, face, registry, prand):
        "Generate a roof for rectangular cross section buildings"
        assert self._isRect (face)
        
        # rearrange verts so that v0-v1 is a short edge
        edges = face.getEdges ()
        lengths_sq = [v.length_sq() for v in edges]
        if (lengths_sq[0] + lengths_sq[2]) > (lengths_sq[1] + lengths_sq[3]):
            verts = face.verts[1:] + face.verts[:1]
            edges = edges[1:] + edges[:1]
        else:
            verts = face.verts

        # work out the apex points
        height = prand.val ('roof-apex-height', 0.2, 0.6) * edges[0].length()
        a0 = verts[0] + (edges[0] * 0.5)
        a0.y += height
        a1 = verts[2] + (edges[2] * 0.5)
        a1.y += height

        # move the apices inwards a bit
        apexshift = prand.val ('roof-apex-shift', 0.1, 0.25)
        ridge = a1 - a0
        a0 += ridge * apexshift
        a1 -= ridge * apexshift
        
        # generate it
        o = Object.New ('roof')
        robj = self.chooseSocketValue ('mat', registry, prand)
        o.material = robj.getObject ()
        self._genTexQuad (o, (verts[1], verts[2], a1, a0))
        self._genTexQuad (o, (verts[3], verts[0], a0, a1))
        self._genTexTri (o, (verts[0], verts[1], a0))
        self._genTexTri (o, (verts[2], verts[3], a1))
	
        return o



def Register ():
    return [ SimpleApex(), SlopedApex() ]

