from PCD import MetaObject, Object, Maths
from PCD.Architecture import Roof
from PCD.Maths.Vector3D import Vec3
import math, operator, random


class FlatRoof (Roof.Func):
    "Walled flat roof"

    def __init__ (self):
        self._regpath = '/roofs'
        self._regname = 'flat-walled'
        self._regsockets = [ ('MAT_WALLFACE', '/materials/wall', True),
                             ('centremat', '/materials/roof/flat'),
                             ('decor', '/objects/decoration/roof/flat') ]


    def __genDecor (self, face, registry, prand):
        
        i = random.randrange (len (face.verts))
        ni = (i+1) % len (face.verts)
        
        e = face.verts[ni] - face.verts[i]
        n = e.cross (face.normal)
        n.normalise()
        el = e.length ()

        obj = self.chooseSocketValue ('decor', registry, prand).getObject ()

        if el < obj.getWidth():
            # we can't add decor here, it's too small
            return Object()
        
        s = random.uniform (0.2, 0.8)
        t = random.uniform (-3.5, -1.5)

        pos = s*e + t*n + face.verts[i]
        yrot = math.atan2 (n.x, n.z)
        
        xform = Maths.Util.MakeXForm (pos, yrot)

        decor = obj.clone ()
        decor.applyTransformFunc (xform)

        return decor
        
        
    def generateContent (self, face, registry, prand):
        
        wall = Object.New ('roofwall')
        robj = self.chooseSocketValue ('MAT_WALLFACE', registry, prand)
        wall.material = robj.getObject ()

        thickness = prand.val ('roof-thickness', 0.22, 0.40)
        height = prand.val ('roof-height', 0.35, 1.0)
    
        # list of outside bottom vertices
        ob = face.verts
        
        # list of outside top vertices
        ot = [v + face.normal*height for v in ob]
    
        # calculate list of inside bottom vertices
        ib = _CalcInnerVerts (ob, face.normal, thickness)
    
        # list of inside top vertices
        it = [v + face.normal*height for v in ib]
    
        # find the centroid
        centroid = reduce (operator.add, ob) / float(len (ob))
            
        # generate the wall
        for i in range (len (ob)):
            ni = (i+1) % len(ob)
    
            w = (ob[ni] - ob[i]).length()
            h = height
            t = thickness
            
            wall.addTri ( (ob[i], ob[ni], ot[ni]),  ((0,0), (w,0), (w,h)) )
            wall.addTri ( (ob[i], ot[ni], ot[i]),   ((0,0), (w,h), (0,h)) )
            wall.addTri ( (ot[i], ot[ni], it[ni]),  ((0,0), (w,0), (w-t,t)) )
            wall.addTri ( (ot[i], it[ni], it[i]),   ((0,0), (w-t,t), (t,t)) )
            wall.addTri ( (it[i], it[ni], ib[ni]),  ((t,0), (w-t,0), (w-t,h)) )
            wall.addTri ( (it[i], ib[ni], ib[i]),   ((t,0), (w-t,h), (t,h)) )
    
        # generate the centre fill
        roof = Object.New ('roofcentre')
        robj = self.chooseSocketValue ('centremat', registry, prand)
        roof.material = robj.getObject ()
        for i in range (len (ob)):
            ni = (i+1) % len(ob)
            t0 = (ib[i].x - centroid.x, ib[i].z - centroid.z)
            t1 = (ib[ni].x - centroid.x, ib[ni].z - centroid.z)
            roof.addTri ( (ib[i], ib[ni], centroid), (t0, t1, (0,0)) )
            
        meta = MetaObject.New ('roof')
        meta.addObjectData (roof)
        meta.addObjectData (wall)
        
        if random.choice ( (True, False, False) ):
            decor = self.__genDecor (face, registry, prand)
            meta.addObjectData (decor)
    
        return meta



def _CalcEdgeNormals (verts, facenormal):
    en = []
    for i in range (len (verts)):
        ni = (i+1) % len (verts)
        
        # work out the "edge" normal
        e = verts[ni] - verts[i]
        n = e.cross (facenormal)
        n.normalise ()

        en.append (n)

    return en
    

def _CalcInnerVerts (verts, facenormal, thickness):
    en = _CalcEdgeNormals (verts, facenormal)
    
    inner = []
    for i in range (len (verts)):
        ni = (i+1) % len (verts)
        pi = (i-1) % len (verts)

        a0 = verts[i] - en[i]*thickness
        d0 = verts[ni] - verts[i]
        a1 = verts[i] - en[pi]*thickness
        d1 = verts[pi] - verts[i]

        da = a1 - a0

        tmp = d1.y*d0.x - d1.x*d0.y
        if math.fabs(tmp) > 0.01:
            t1 = (d0.y*da.x - d0.x*da.y) / (d1.y*d0.x - d1.x*d0.y)
        else:
            t1 = (d0.z*da.x - d0.x*da.z) / (d1.z*d0.x - d1.x*d0.z)

        inner.append (a1 + d1*t1)

    return inner
    


def Register ():
    return [ FlatRoof() ]


